export class RingBuffer {

  constructor() {
    this._array = [];
  }

  unshift(item) {
    this._array.unshift(item);
  }

  pop() {
    return this._array.pop();
  }

  count() {
    return this._array.length;
  }

  cleanup(predicate) {
    this._array = this._array.filter(predicate);
  }
}

export class FixedBuffer {

  constructor(buffer, n) {
    this._buffer = buffer;
    this._n = n;
  }

  isFull() {
    return this._buffer.count() >= this._n;
  }

  add(item) {
    this._buffer.unshift(item);
  }

  remove() {
    return this._buffer.pop();
  }

  count() {
    return this._buffer.count();
  }
}

export class DroppingBuffer {

  constructor(buffer, n) {
    this._buffer = buffer;
    this._n = n;
  }

  isFull() {
    return false;
  }

  add(item) {
    if (this._buffer.count() < this._n) {
      this._buffer.unshift(item);
    }
  }

  remove() {
    return this._buffer.pop();
  }

  count() {
    return this._buffer.count();
  }
}

export class SlidingBuffer {

  constructor(buffer, n) {
    this._buffer = buffer;
    this._n = n;
  }

  isFull() {
    return false;
  }

  add(item) {
    if (this._buffer.count() === this._n) {
      this._buffer.pop();
    }

    this._buffer.unshift(item);
  }

  remove() {
    return this._buffer.pop();
  }

  count() {
    return this._buffer.count();
  }
}

/**
 * Returns a buffer that is considered "full" when it reaches size n,
 * but still accepts additional items, effectively allow overflowing.
 * The overflowing behavior is useful for supporting "expanding"
 * transducers, where we want to check if a buffer is full before
 * running the transduced step function, while still allowing a
 * transduced step to expand into multiple "essence" steps.
 */


export function ring() {
  return new RingBuffer();
}

export function fixed(n) {
  return new FixedBuffer(ring(), n);
}

export function dropping(n) {
  return new DroppingBuffer(ring(), n);
}

export function sliding(n) {
  return new SlidingBuffer(ring(), n);
}